#include "fm.h"

/*
 * _fm_init_state()
 *
 * Initialize a state object.
 */
void _fm_init_state(fm_state_t *state)
{
    state->state = FM_STATE_UNDEFINED;
    nexus_mutex_init(&state->lock, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&state->cond, (nexus_condattr_t *) NULL);
#ifdef FM_DEBUG
    if (FMDebug(State, 2))
	nexus_printf("_fm_init_state(): state=%x\n",
		     (unsigned long) state);
#endif
} /* _fm_init_state */


/*
 * _fm_set_state()
 *
 * Set a state object to a value.
 */
void _fm_set_state(fm_state_t *state, int new_value)
{
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_set_state(): state=%x new_value=%d waiting for lock\n",
		     (unsigned long) state, new_value);
#endif
    nexus_mutex_lock(&state->lock);
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_set_state(): state=%x new_value=%d old_value=%d\n",
		     (unsigned long) state, new_value, state->state);
#endif
    state->state = new_value;
    nexus_cond_signal(&state->cond);
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_set_state(): state=%x unlocking lock\n",
		     (unsigned long) state);
#endif
    nexus_mutex_unlock(&state->lock);
} /* _fm_set_state */


/*
 * _fm_set_locked_state()
 *
 * Set a state object to a value, assuming it is already locked.
 */
void _fm_set_locked_state(fm_state_t *state, int new_value)
{
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_set_locked_state(): state=%x new_value=%d old_value=%d\n",
		     (unsigned long) state, new_value, state->state);
#endif
    state->state = new_value;
} /* _fm_set_locked_state */


/*
 * _fm_state()
 * 
 * Return the current value of the state.
 * 
 */
int _fm_state(fm_state_t *state)
{
    return state->state;
    
} /* _fm_state() */


/*
 * _fm_lock_state()
 * 
 * Lock the state.
 * 
 */
void _fm_lock_state(fm_state_t *state)
{
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_lock_state(): state=%x waiting for lock\n",
		     (unsigned long) state);
#endif
    nexus_mutex_lock(&state->lock);
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_lock_state(): state=%x got lock\n",
		     (unsigned long) state);
#endif
} /* _fm_lock_state() */


/*
 * _fm_unlock_state()
 * 
 * Unlock the state.
 * 
 */
void _fm_unlock_state(fm_state_t *state)
{
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_unlock_state(): state=%x unlocking lock\n",
		     (unsigned long) state);
#endif
    nexus_mutex_unlock(&state->lock);
} /* _fm_lock_state() */




/*
 * _fm_wait_state()
 *
 * Wait for the state to take on the given value.
 */
void _fm_wait_state(fm_state_t *state, int desired_value)
{
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_wait_state(): state=%x desired_value=%d waiting for lock\n",
		     (unsigned long) state, desired_value);
#endif
    nexus_mutex_lock(&state->lock);
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_wait_state(): state=%x got lock value=%d desired=%d\n",
		     (unsigned long) state, state->state, desired_value);
#endif
    while (state->state != desired_value)
    {
#ifdef FM_DEBUG
	if (FMDebug(State, 1))
	    nexus_printf("_fm_wait_state(): state=%x in loop value=%d desired=%d\n",
			 (unsigned long) state, state->state, desired_value);
#endif
	nexus_cond_wait(&state->cond, &state->lock);
    }
#ifdef FM_DEBUG
    if (FMDebug(State, 1))
	nexus_printf("_fm_wait_state(): state=%x unlocking value=%d\n",
		     (unsigned long) state, state->state);
#endif
    nexus_mutex_unlock(&state->lock);
	
} /* _fm_wait_state */


