/**********************************************************************
$Id: port_io.c,v 1.12 1995/04/17 06:30:23 kohr Exp $

$Source: /home/compose/Master/FM/src/fm_runtime/port_io.c,v $

$Log: port_io.c,v $
Revision 1.12  1995/04/17  06:30:23  kohr
Doc. added, moved migration functions to migrate.c.

Revision 1.11  1995/04/06  06:32:27  kohr
Added documentation, cleaned up a bit.


FM runtime routines for putting and getting ports through Nexus buffers.
**********************************************************************/


#include "fm.h"

/*
We pass ports over channels using a global pointer to a port that ALREADY
EXISTS on the destination. The code in this module is not responsible for
enforcing this restriction. That responsibility is that of the process
block or port migration code.
*/

/*
_fm_sizeof_inport() : An inport is just passed as the global pointer to the
future home.
*/

int _fm_sizeof_inport(nexus_buffer_t *buffer,
		      f_integer_t *f_inports,
		      int n_inports,
		      int *n_elts)
{
    fm_inport_t *inport;
    int size, obj_n_elts;
    int i;

    size = *n_elts = 0;

    for (i = 0; i < n_inports; i++)
    {
	inport = (fm_inport_t *) f_inports[i];
	InportMagicCheck(inport);
	NexusAssert(nexus_global_pointer_to_current_context(
		&inport->channel_gp));

	size += nexus_sizeof_global_pointer(buffer, &inport->future_home, 1,
					    &obj_n_elts);
	*n_elts += obj_n_elts;
    }

    return size;
}	      


/*
_fm_put_inport() : Put an inport onto the buffer. It is simply passed as a
global pointer to the future home.
*/

void _fm_put_inport(nexus_buffer_t *buffer,
		     f_integer_t *f_inports,
		     int n_inports)
{
    int i;
    fm_inport_t *inport;
    

    for (i = 0; i < n_inports; i++)
    {
	inport = (fm_inport_t *) f_inports[i];
	InportMagicCheck(inport);

	NexusAssert(nexus_global_pointer_to_current_context(
		&inport->channel_gp));
	nexus_put_global_pointer(buffer, &inport->future_home, 1);
    }
    
} /* _fm_put_inport() */


/*
_fm_get_inport()
*/

void _fm_get_inport(nexus_buffer_t *buffer,
		    f_integer_t *inports,
		    int n_inports,
		    void (*init_func)(fm_channel_t *))
{
    nexus_global_pointer_t gp;
    int i;
    fm_inport_t *inport;
    fm_channel_t *channel;

    for (i = 0; i < n_inports; i++)
    {
        nexus_get_global_pointer(buffer, &gp, 1);
#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_inport(): Got inport %s\n",
	    		 nexus_global_pointer_string(&gp));
#endif
        NexusAssert(nexus_global_pointer_to_current_context(&gp));
        inport = (fm_inport_t *) nexus_convert_global_pointer_address(&gp);
        InportMagicCheck(inport);

	InportLock(inport);

#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_inport(): Got inport %s\n",
			 _fm_inport_string(inport));
#endif

	/*
	 * Grab the channel pointer and set invoke the init func
	 * to set up the function pointers appropriately.
	 */

	NexusAssert(nexus_global_pointer_to_current_context(
		&inport->channel_gp));
	channel = (fm_channel_t *)
	    nexus_convert_global_pointer_address(&inport->channel_gp);
	ChannelMagicCheck(channel);

	ChannelLock(channel);
	
#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_inport(): invoking init_func %x on "
	    		 "channel %s\n",
			 (unsigned long) init_func,
			 _fm_channel_string(channel));
#endif

	NexusAssert(init_func != (void (*)(fm_channel_t*)) NULL);
	(*init_func)(channel);

	ChannelSetState(channel, FM_STATE_NORMAL);
	InportSetState(inport, FM_STATE_NORMAL);

	ChannelUnlock(channel);
	InportUnlock(inport);
	
        inports[i] = (f_integer_t) inport;
    }
} /* _fm_get_inport() */
                    

/*
_fm_sizeof_outport() : Compute the size of an outport.
*/

int _fm_sizeof_outport(nexus_buffer_t *buffer,
		      f_integer_t *f_outports,
		      int n_outports,
		      int *n_elts)
{
    int size, obj_n_elts;
    int i;
    fm_outport_t *outport;

    size = *n_elts = 0;

    for (i = 0; i < n_outports; i++)
    {
	outport = (fm_outport_t *) f_outports[i];

	OutportMagicCheck(outport);
	
	size += nexus_sizeof_global_pointer(buffer,
					    &outport->future_home,
					    1,
					    &obj_n_elts);
	*n_elts += obj_n_elts;

    }
    return size;
}	      


/*
_fm_put_outport() : Stuff an outport into a message buffer.
*/

void _fm_put_outport(nexus_buffer_t *buffer,
		     f_integer_t *f_outports,
		     int n_outports)
{
    int i;
    fm_outport_t *outport;

    for (i = 0; i < n_outports; i++)
    {
	outport = (fm_outport_t *) f_outports[i];

	NexusAssert(!nexus_is_null_global_pointer(&outport->future_home));
	nexus_put_global_pointer(buffer, &outport->future_home, 1);
    }
}	      


/*
_fm_get_outport()
*/

void _fm_get_outport(nexus_buffer_t *buffer,
                     f_integer_t *outports,
                     int n_outports)
{
    nexus_global_pointer_t gp;
    int i;
    fm_outport_t *outport;

    for (i = 0; i < n_outports; i++)
    {
        nexus_get_global_pointer(buffer, &gp, 1);

        NexusAssert(nexus_global_pointer_to_current_context(&gp));
        outport = (fm_outport_t *) nexus_convert_global_pointer_address(&gp);
        OutportMagicCheck(outport);
	
#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_outport(): Got outport %s\n",
			 _fm_outport_string(outport));
#endif
	
        outports[i] = (f_integer_t) outport;
    }
} /* _fm_get_outport() */
                    

/*
_fm_get_stashed_outport()
*/

void _fm_get_stashed_outport(nexus_stashed_buffer_t *buffer,
			     f_integer_t *outports,
			     int n_outports)
{
    nexus_global_pointer_t gp;
    int i;
    fm_outport_t *outport;

    for (i = 0; i < n_outports; i++)
    {
        nexus_get_stashed_global_pointer(buffer, &gp, 1);

        NexusAssert(nexus_global_pointer_to_current_context(&gp));
        outport = (fm_outport_t *) nexus_convert_global_pointer_address(&gp);
        OutportMagicCheck(outport);
	
#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_stashed_outport(): Got outport %s\n",
			 _fm_outport_string(outport));
#endif
	
        outports[i] = (f_integer_t) outport;
    }
} /* _fm_get_stashed_outport() */


/*
_fm_get_stashed_inport()
*/

void _fm_get_stashed_inport(nexus_stashed_buffer_t *buffer,
			    f_integer_t *inports,
			    int n_inports,
			    void (*init_func)(fm_channel_t *))
{
    nexus_global_pointer_t gp;
    int i;
    fm_inport_t *inport;
    fm_channel_t *channel;

    for (i = 0; i < n_inports; i++)
    {
        nexus_get_stashed_global_pointer(buffer, &gp, 1);
#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_stashed_inport(): Got inport %s\n",
	    		 nexus_global_pointer_string(&gp));
#endif
        NexusAssert(nexus_global_pointer_to_current_context(&gp));
        inport = (fm_inport_t *) nexus_convert_global_pointer_address(&gp);
        InportMagicCheck(inport);

	InportLock(inport);

#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_stashed_inport(): Got inport %s\n",
			 _fm_inport_string(inport));
#endif

	/*
	 * Grab the channel pointer and set invoke the init func
	 * to set up the function pointers appropriately.
	 */

	NexusAssert(nexus_global_pointer_to_current_context(
		&inport->channel_gp));
	channel = (fm_channel_t *)
	    nexus_convert_global_pointer_address(&inport->channel_gp);
	ChannelMagicCheck(channel);

	ChannelLock(channel);
	
#ifdef FM_DEBUG
	if (FMDebug(Ports, 1))
	    nexus_printf("_fm_get_stashed_inport(): invoking init_func %x "
	    		 "on channel %s\n",
			 (unsigned long) init_func,
			 _fm_channel_string(channel));
#endif

	NexusAssert(init_func != (void (*)(fm_channel_t*)) NULL);
	(*init_func)(channel);

	ChannelSetState(channel, FM_STATE_NORMAL);
	InportSetState(inport, FM_STATE_NORMAL);

	ChannelUnlock(channel);
	InportUnlock(inport);
	
        inports[i] = (f_integer_t) inport;
    }
} /* _fm_get_stashed_inport() */
