
/* #define PRINT_ALL /* */

#include "fm_config.h"

#ifdef FM_SENTINEL_SUPPORT

#define FM_FREELIST_DONT_PREPEND_FM
#define FM_PORT_ALLOC_SIZE 64

#include "freelist.h"
#include "fortran_types.h"
#include "nexus.h"
#include "sentinel.h"

/*
 * Sentinel debugging support.
 * 
 * The idea is to keep for each thread a stack of "sentinels", local 
 * variables with known values. Every time a subprogram begins execution, 
 * a new sentinel is pushed onto the stack. When it exits, it pops the
 * sentinel stack. 
 * 
 * At arbitrary times (say for every call into the sentinel library) 
 * the stack is scanned to ensure that all sentinels are still valid. 
 * If they are not, the program aborts with an error message.
 * 
 * There is also support for "static" sentinels for things that live 
 * in common blocks. This is useful for things like FM ports that 
 * change values without warning :-).
 *
 */

#ifdef FM_NEED_TRAILING_UNDERSCORES

#define SENTINEL_PUSH sentinel_push_
#define SENTINEL_POP sentinel_pop_
#define SENTINEL_SCAN sentinel_scan_

#else /* FM_NEED_TRAILING_UNDERSCORES */

#ifndef FM_NEED_UPPERCASE

#define SENTINEL_PUSH sentinel_push
#define SENTINEL_POP sentinel_pop
#define SENTINEL_SCAN sentinel_scan

#endif /* !FM_NEED_UPPERCASE */

#endif /* FM_NEED_TRAILING_UNDERSCORES */

#define SENTINEL_MAGIC 0x91827364

typedef struct _sentinel_stack_entry_t
{
    void *address;
    int value;
    char *name;
    int depth;
    struct _sentinel_stack_entry_t *next;
} sentinel_stack_entry_t;

typedef struct _sentinel_stack_t
{
    sentinel_stack_entry_t *stack;
    int cur_depth;
    sentinel_stack_entry_t *static_entries;
} sentinel_stack_t;

static nexus_thread_key_t sentinel_key;

#define GET_STACK							      \
    stack = (sentinel_stack_t *) nexus_thread_getspecific(sentinel_key);      \
    if (stack == (sentinel_stack_t *) NULL)				      \
    {									      \
	stack = alloc_sentinel_stack();					      \
	nexus_thread_setspecific(sentinel_key, (void *) stack);		      \
    }

static void init_sentinel_stack_entry(sentinel_stack_entry_t *ent)
{
    ent->address = (void *) NULL;
    ent->value = -1;
    ent->name = "";
    ent->depth = -1;
    ent->next = (sentinel_stack_entry_t *) NULL;
}

static void reinit_sentinel_stack_entry(sentinel_stack_entry_t *ent)
{
    ent->address = (void *) NULL;
    ent->value = -1;
    ent->name = "";
    ent->depth = -1;
    ent->next = (sentinel_stack_entry_t *) NULL;
}

static sentinel_stack_t *alloc_sentinel_stack()
{
    sentinel_stack_t *stack;

    NexusMalloc(alloc_sentinel_stack(), stack, sentinel_stack_t *,
		sizeof(sentinel_stack_t));
    stack->stack = (sentinel_stack_entry_t *) NULL;
    stack->static_entries = (sentinel_stack_entry_t *) NULL;
    stack->cur_depth = 0;
    return stack;
}

FM_FREE_LIST_CODE(sentinel_stack_entry)

void sentinel_init()
{
    nexus_thread_key_create(&sentinel_key,
			    (nexus_thread_key_destructor_func_t) NULL);

    FM_FREE_LIST_INIT(sentinel_stack_entry);
}

void SENTINEL_PUSH(f_integer_t *sentinel_addr,
		   f_character_t subprogram_name,
		   f_integer_t *first_var,
		   f_integer_t *last_var
		   CHAR_LEN_DECL(subprogram_name_len))
{
    LOCAL_CHAR_LEN_DECL(subprogram_name_len)
    char *name;

    CONVERT_CHAR_FTN_TO_C_MALLOC(subprogram_name, subprogram_name_len, name);

    *sentinel_addr = SENTINEL_MAGIC;

    sentinel_push_value((void *) sentinel_addr,
			SENTINEL_MAGIC,
			name);

    sentinel_scan(name, "push");
}

void SENTINEL_POP(f_integer_t *sentinel_addr,
		  f_character_t subprogram_name,
		  f_integer_t *first_var,
		  f_integer_t *last_var
		  CHAR_LEN_DECL(subprogram_name_len))
{
    LOCAL_CHAR_LEN_DECL(subprogram_name_len)
    char name[1024];

    CONVERT_CHAR_FTN_TO_C_INPLACE(subprogram_name, subprogram_name_len, name);

    sentinel_scan(name, "pop");

    sentinel_pop_value((void *) sentinel_addr,
		       (int) *sentinel_addr,
		       name);
}

void sentinel_push_value(void *addr, int value, char *name)
{
    sentinel_stack_entry_t *ent;
    sentinel_stack_t *stack;

    stack = (sentinel_stack_t *) nexus_thread_getspecific(sentinel_key);
    if (stack == (sentinel_stack_t *) NULL)				      
    {									      
	stack = alloc_sentinel_stack();					      
	nexus_thread_setspecific(sentinel_key, (void *) stack);		      
    }
    
    ent = _fm_alloc_sentinel_stack_entry();

    ent->address = addr;
    ent->value = value;
    ent->name = name;
    ent->depth = stack->cur_depth++;

    ent->next = stack->stack;
    stack->stack = ent;

#ifdef PRINT_ALL
    nexus_printf("sentinel: pushed addr=%x value=%d name=%s depth=%d\n",
		 (unsigned long) addr, value, name, ent->depth);
#endif
}

void sentinel_pop_value(void *addr, int value, char *name)
{
    sentinel_stack_entry_t *ent, *nent, *newstack;
    sentinel_stack_t *stack;
    int npop;
    int depth;

    GET_STACK;

    npop = 0;
    for (ent = stack->stack; ent != (sentinel_stack_entry_t *) NULL;
	 ent = ent->next)
    {
	if (ent->address == addr)
	    break;
	nexus_warning("sentinel_pop_value(%s %x): Popping frame addr=%x value=%x name=%s depth=%d\n",
		      name, (unsigned long) addr,
		      (unsigned long) ent->address,
		      ent->value, ent->name, ent->depth);
	npop++;
    }

    if (ent == (sentinel_stack_entry_t *) NULL)
    {
	nexus_warning("sentinel_pop_value(%s %x): Entry not found\n",
		      name,
		      (unsigned long) addr);
	return;
    }

    if (npop > 0)
    {
	nexus_warning("sentinel_pop_value(%s %x): Had to pop %d extra frames\n",
		      name,
		      (unsigned long) addr,
		      npop);
    }

    depth = ent->depth;

    newstack = ent->next;
    while (stack->stack != newstack)
    {
	ent = stack->stack;
	stack->stack = stack->stack->next;
	
	_fm_free_sentinel_stack_entry(ent);
    }

    stack->cur_depth -= (npop + 1);
#ifdef PRINT_ALL
    nexus_printf("sentinel: popped addr=%x value=%d name=%s depth=%d cur_depth=%d\n",
		 (unsigned long) addr, value, name, depth, stack->cur_depth);
#endif

}

void sentinel_scan(char *name, char *where)
{
    sentinel_stack_entry_t *ent;
    sentinel_stack_t *stack;
    int new;

    GET_STACK;

    for (ent = stack->stack; ent != (sentinel_stack_entry_t *) NULL;
	 ent = ent->next)
    {
	new = *((int *) ent->address);
	if (ent->value != new)
	{
	    nexus_warning("SENTINEL CHANGED: addr=%x old_value=%x new_value=%x name=%s depth=%d cur_depth=%d (while scanning from %s %s)\n",
			  (unsigned long) ent->address,
			  ent->value,
			  new,
			  ent->name,
			  ent->depth,
			  stack->cur_depth,
			  name, where);
	    ent->value = new;
	}
    }
    if (getenv("DO_SCAN_PRINT"))
    {
	static int count = 0;
	nexus_printf("scan ok %d\n", count);
	count++;
    }
/*    MA_verify_allocator_stuff();*/
}

#ifdef FM_NEED_TRAILING_UNDERSCORES

void sentinel_scan_(char *name)
{
    sentinel_scan(name, "");
}

#endif

/*
 * FM interface routines
 */

void sentinel_args_init(int *argc, char ***argv)
{
}

void sentinel_usage_message()
{
    printf("Sentinel support included!\n");
}

int sentinel_params(int n, char *buf, int size)
{
    return n;
}


/*
 * Routines for the pr_tcp module
 */

void sentinel_push_tcp(int *addr, int line)
{
    char *s;

    NexusMalloc(sentinel_push_tcp(), s, char *, 10);
    sprintf(s, "%d", line);
    *addr = SENTINEL_MAGIC;
    sentinel_push_value((void *) addr, *addr, s);
    sentinel_scan(s, "push");
}

void sentinel_pop_tcp(int *addr, int line)
{
    char s[10];

    sprintf(s, "%d", line);
    sentinel_scan(s, "pop");
    sentinel_pop_value((void *) addr, *addr, s);
}

#endif /* FM_SENTINEL_SUPPORT */
