/*
 * pr_iface.c
 *
 * Protocol interface routines
 */

static char *rcsid = "$Header: /home/compose/Master/FM/src/nexus/pr_iface.c,v 1.30 1995/08/22 20:31:59 tuecke Exp $";
static char *nexus_rcsid = "Header: /Net/folding.compbio.caltech.edu/bcomp/src/master/nexus/pr_iface.c,v 1.46 1995/08/21 23:39:26 patton Exp ";

#include "internal.h"

#include <netinet/in.h>
#ifdef NEXUS_ARCH_AXPOSF
#define NX_NTOHL(m) m
#define NX_HTONL(m) m
#else
#define NX_NTOHL(m) ntohl(m)
#define NX_HTONL(m) htonl(m)
#endif

static void free_mi_proto(nexus_mi_proto_t *mi_proto);

#ifdef NEXUS_DEBUG
/* These are used by nexus_global_pointer_string() */
#define GP_N_STRINGS 10
static nexus_mutex_t gp_string_mutex;
#endif /* NEXUS_DEBUG */

#ifdef NEXUS_PROFILE
#define NEXUS_POLL_RANGE_CHECK
static nexus_bool_t check_poll_ranges;
static nexus_bool_t check_poll_traceback;
static double last_poll_time;
static double poll_range_min;
static double poll_range_max;
#endif

static void _nx_copy_mi_proto( nexus_mi_proto_t *mi_proto_in,
			       nexus_mi_proto_t **mi_proto_out);

/*
 * Proto module list
 */
typedef struct _proto_module_list_t
{
    nexus_proto_funcs_t *		funcs;
    char *				name;
    nexus_proto_type_t			type;
    struct _proto_module_list_t *	next;
} proto_module_list_t;
static proto_module_list_t *	proto_module_list_head;
static proto_module_list_t *	proto_module_list_tail;

#define AddProtoModuleToList(Caller, Name, Funcs, Type) \
{ \
    proto_module_list_t *__p; \
    NexusMalloc(Caller, __p, proto_module_list_t *, \
		sizeof(proto_module_list_t)); \
    __p->name = _nx_copy_string(Name); \
    __p->funcs = (Funcs); \
    __p->type = (Type); \
    __p->next = (proto_module_list_t *) NULL; \
    if (proto_module_list_head) \
    { \
	proto_module_list_tail->next = __p; \
	proto_module_list_tail = __p; \
    } \
    else \
    { \
	proto_module_list_head = proto_module_list_tail = __p; \
    } \
}

static int			n_protos_with_poll;
static proto_module_list_t *	proto_module_with_poll;

static nexus_mi_proto_t *	my_mi_protos;

/*
 * saved proto arguments
 */
static char saved_no_proto_arg[256];


/*
 * _nx_proto_usage_message()
 *
 * Call the usage_message() function for each protocol
 */
void _nx_proto_usage_message()
{
#ifdef NEXUS_POLL_RANGE_CHECK    
    printf("    -poll_check <min> <max>   : Print a notice if the time between\n");
    printf("                                nexus_poll() calls does not fall in the\n");
    printf("                                supplied range. (<min> and <max> are reals)\n");
    printf("    -poll_trace               : Print stack tracebacks with\n");
    printf("                                poll_check messages\n");
#endif /* NEXUS_POLL_RANGE_CHECK */	
} /* _nx_proto_usage_message() */


/*
 * _nx_proto_new_process_params()
 *
 * Call the new_process_params() function for each protocol
 *
 * Each of those functions may add stuff to 'buf', returning the number
 * of characters that they added.
 *
 * Return: The total number of characters added to 'buf'.
 */
int _nx_proto_new_process_params(char *buf, int size)
{
    char tmp_buf[256];
    int n_left = size;
    char *b = buf;
    int n_added;

    n_added = 0;
    tmp_buf[0] = '\0';

    if (strlen(saved_no_proto_arg) > 0)
    {
	char tmp_buf2[256];

	sprintf(tmp_buf2, "-no_pr %s ", saved_no_proto_arg);
	strcat(tmp_buf, tmp_buf2);
    }
#ifdef NEXUS_POLL_RANGE_CHECK
    if (check_poll_ranges)
    {
	char tmp_buf2[256];

	sprintf(tmp_buf2, "-poll_check %f %f ",
		poll_range_min, poll_range_max);
	if (check_poll_traceback)
	{
	    strcat(tmp_buf2, "-poll_trace ");
	}
	strcat(tmp_buf, tmp_buf2);
    }
#endif /* NEXUS_POLL_RANGE_CHECK */	
    n_added = strlen(tmp_buf);
    if (n_added > n_left)
    {
        nexus_fatal("_nx_proto_new_process_params(): Internal error: Not enough room in buffer for arguments\n");
    }
    strcpy(b, tmp_buf);
    b += n_added;
    n_left -= n_added;

    return (size - n_left);
} /* _nx_proto_new_process_params() */


/*
 * _nx_proto_init()
 *
 * Initialize the protocol modules.
 *
 * TODO: This needs to be made more dynamic
 */
void _nx_proto_init(int *argc, char ***argv,
		    nexus_module_list_t module_list[])
{
    int i;
    int rc;
    nexus_proto_funcs_t *proto_funcs;
    nexus_proto_type_t proto_type;
    proto_module_list_t *proto_module;
    nexus_mi_proto_t *last_mi_proto, *mi_proto;
#ifndef JGG
#else
    char *all_proto_names;
    char *value, *next;
#endif

    int arg_num;
    char *arg;

    char *next, *no_proto;
    nexus_bool_t add_proto;


    /*
     * Parse any arguments
     */
    if ((arg_num = nexus_find_argument(argc, argv, "no_pr", 2)) >= 0)
    {
	arg = (*argv)[arg_num + 1];
	strcpy(saved_no_proto_arg, arg);
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    
#ifdef NEXUS_POLL_RANGE_CHECK
    check_poll_ranges = NEXUS_FALSE;
    check_poll_traceback = NEXUS_FALSE;
    last_poll_time = -1.0;
    if ((arg_num = nexus_find_argument(argc, argv, "poll_check", 3)) >= 0)
    {
	check_poll_ranges = NEXUS_TRUE;
	poll_range_min = atof((*argv)[arg_num + 1]);
	poll_range_max = atof((*argv)[arg_num + 2]);
	nexus_remove_arguments(argc, argv, arg_num, 3);
    }
    if ((arg_num = nexus_find_argument(argc, argv, "poll_trace", 1)) >= 0)
    {
	check_poll_traceback = NEXUS_TRUE;
	nexus_remove_arguments(argc, argv, arg_num, 1);
    }
#endif /* NEXUS_POLL_RANGE_CHECK */

#ifdef NEXUS_DEBUG
    nexus_mutex_init(&gp_string_mutex, (nexus_mutexattr_t *) NULL);
#endif

    /*
     * Scan the module_list looking for protocol modules.
     * For each of these, get the function table,
     * and add that module to the proto_module_list.
     * Put the local protocol module at the front of the list.
     */
    proto_module_list_head = proto_module_list_tail
	= (proto_module_list_t *) NULL;
    proto_funcs = _nx_pr_local_info();
    proto_type = (*proto_funcs->proto_type)();
    AddProtoModuleToList(_nx_proto_init(),
			 "local",
			 proto_funcs,
			 proto_type);

    for (i = 0; module_list[i].family_name != (char *) NULL; i++)
    {
	if ((strcmp(module_list[i].family_name, "protocols") == 0))
	{
	    add_proto = NEXUS_TRUE;
	    next = saved_no_proto_arg;
	    /* 
	     * This can be changed to while(next && add_proto) if the
	     * number of no_pr options increase.  With only tcp, atm,
	     * and inx, this is not an optimization.
	     */
	    while (next)
	    {
		_nx_get_next_value(next, ':', &next, &no_proto);
		if (strcmp(module_list[i].module_name, no_proto) == 0)
		{
		    add_proto = NEXUS_FALSE;
		}
	    }
	    /*
	    rc = _nx_module_load(module_list[i].family_name,
				 module_list[i].module_name,
				 (void *) &proto_funcs);
	    */
	    rc = 0;
	    if (rc == 0 && add_proto)
	    {
	        proto_funcs
		    = (nexus_proto_funcs_t *) (*module_list[i].info_func)();
		proto_type = (*proto_funcs->proto_type)();
		AddProtoModuleToList(_nx_proto_init(),
				     module_list[i].module_name,
				     proto_funcs,
				     proto_type);
	    }
	}
    }

    /*
     * Initialize each protocol module
     */
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	(*proto_module->funcs->init)(argc, argv);
    }

    /*
     * Find out how many of the protos have poll functions.
     */
    n_protos_with_poll = 0;
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->poll)
	{
	    n_protos_with_poll++;
	    proto_module_with_poll = proto_module;
	}
    }
    
    /*
     * Get my nexus_mi_proto_t list for each module, for use
     * in global pointers that I create.
     */
    last_mi_proto = (nexus_mi_proto_t *) NULL;
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->get_my_mi_proto)
	{
	    mi_proto = (proto_module->funcs->get_my_mi_proto)();
	    if (mi_proto)
	    {
		if (last_mi_proto)
		{
		    last_mi_proto->next = mi_proto;
		}
		else
		{
		    my_mi_protos = mi_proto;
		}
		mi_proto->next = (nexus_mi_proto_t *) NULL;
		last_mi_proto = mi_proto;
	    }
	}
    }
} /* _nx_proto_init() */


/*
 * _nx_proto_shutdown()
 *
 * If 'shutdown_others' is NEXUS_TRUE,
 * then try to shutdown all other nodes and contexts.
 */
void _nx_proto_shutdown(nexus_bool_t shutdown_others)
{
    proto_module_list_t *proto_module;
    
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->shutdown)
	{
	    (proto_module->funcs->shutdown)(shutdown_others);
	}
    }
} /* _nx_proto_shutdown() */


/*
 * _nx_proto_abort()
 */
void _nx_proto_abort()
{
    proto_module_list_t *proto_module;
    
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->abort)
	{
	    (proto_module->funcs->abort)();
	}
    }
} /* _nx_proto_abort() */


/*
 * nexus_poll
 *
 * Poll each proto, so that they can handle any outstanding remote
 * service requests.
 *
 * No need to poll the local protocol.
 */
void nexus_poll()
{
    proto_module_list_t *proto_module;
    
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (proto_module->funcs->poll)
	{
	    (proto_module->funcs->poll)();
	}
    }
} /* nexus_poll() */


/*
 * nexus_poll_profile
 *
 * Poll each proto, so that they can handle any outstanding remote
 * service requests.
 *
 * No need to poll the local protocol.
 */
void nexus_poll_profile(int id)
{
    
#ifdef NEXUS_POLL_RANGE_CHECK
    if (check_poll_ranges && last_poll_time >= 0)
    {
	double time_since_last_poll;
	time_since_last_poll = nexus_wallclock() - last_poll_time;
	if (time_since_last_poll < poll_range_min)
	{
	    nexus_notice("poll id %d: Time since last nexus_poll() (%f) is less than minimum\n",
			 id, time_since_last_poll);
	    if (check_poll_traceback)
	    {
		_nx_traceback();
	    }
	}
	else if (time_since_last_poll > poll_range_max)
	{
	    nexus_notice("poll id %d: Time since last nexus_poll() (%f) exceeds maximum\n",
			 id, time_since_last_poll);
	    if (check_poll_traceback)
	    {
		_nx_traceback();
	    }
	}
    }
#endif /* NEXUS_POLL_RANGE_CHECK */    

    nexus_poll();
    
#ifdef NEXUS_POLL_RANGE_CHECK
    if (check_poll_ranges)
    {
	last_poll_time = nexus_wallclock();
    }
#endif /* NEXUS_POLL_RANGE_CHECK */    
    
} /* nexus_poll_profile() */


/*
 * nexus_blocking_poll()
 *
 * If only one proto has a poll routine, then call
 * its blocking_poll() routine.
 * If more than one proto has a poll routine, then call
 * each of those poll() routines.
 *
 * No need to poll the local protocol.
 */
void nexus_blocking_poll()
{
    if (n_protos_with_poll == 1)
    {
	(*proto_module_with_poll->funcs->blocking_poll)();
    }
    else if (n_protos_with_poll > 1)
    {
	nexus_poll();
    }
} /* nexus_blocking_poll() */


/*
 * _nx_must_poll()
 *
 * Return NEXUS_TRUE if nexus_poll() and/or nexus_blocking_poll()
 * should be used -- that is,
 * are there one or more protocol modules with poll routines.
 * Return NEXUS_FALSE otherwise.
 */
nexus_bool_t _nx_must_poll()
{
    if (n_protos_with_poll >= 1)
	return (NEXUS_TRUE);
    else
	return (NEXUS_FALSE);
} /* _nx_must_poll() */


/*
 * nexus_global_pointer()
 */
void nexus_global_pointer(nexus_global_pointer_t *new_gp,
			  void *address)
{
    nexus_context_t *context;
    _nx_context(&context);

    new_gp->proto = _nx_local_proto;
    new_gp->all_protos = my_mi_protos;
    new_gp->context = (unsigned long) context;
    new_gp->address = (unsigned long) address;
    new_gp->class_type = NEXUS_CLASS_TYPE;
#ifdef NEXUS_PROFILE
    _nx_node_id(&(new_gp->node_id));
    _nx_context_id(&(new_gp->context_id));
#endif /* NEXUS_PROFILE */    
} /* nexus_global_pointer() */


/*
 * nexus_convert_global_pointer_address()
 */
void *nexus_convert_global_pointer_address(nexus_global_pointer_t *gp)
{
#ifdef NEXUS_DEBUG
    nexus_context_t *context;
    _nx_context(&context);
    NexusAssert2((((unsigned long) context) == gp->context),
		 ("nexus_convert_global_pointer_address(): Global pointer is not for this context\n") );
#endif
    
    return ((void *) (gp->address));
} /* nexus_convert_global_pointer_address() */


/*
 * nexus_destroy_global_pointer()
 */
void nexus_destroy_global_pointer(nexus_global_pointer_t *gp)
{
    nexus_mi_proto_t *mi_proto, *tmp_mi_proto;

    if (gp->proto)
    {
	(gp->proto->funcs->destroy_proto)(gp->proto);
    }

    if (gp->all_protos != my_mi_protos)
    {
	mi_proto = gp->all_protos;
	while (mi_proto)
	{
	    tmp_mi_proto = mi_proto->next;
	    free_mi_proto(mi_proto);
	    mi_proto = tmp_mi_proto;
	}
    }

    nexus_macro_null_global_pointer(gp);
} /* nexus_destroy_global_pointer() */

/*
 * nexus_copy_global_pointer()
 *
 * Copies the global pointer gp_in to gp_out allowing for
 * reference counting in the protocol modules
 * Copies the context/address then calls copy_proto
 * and _nx_copy_mi_proto()
 */
void nexus_copy_global_pointer( nexus_global_pointer_t *gp_out,
			   nexus_global_pointer_t *gp_in)
{
    gp_out->context = gp_in->context;
    gp_out->address = gp_in->address;
    gp_out->class_type = gp_in->class_type;
#ifdef NEXUS_PROFILE
    gp_out->node_id = gp_in->node_id;
    gp_out->context_id = gp_in->context_id;
#endif /* NEXUS_PROFILE */
    if (gp_in->proto)
    {
	(gp_in->proto->funcs->copy_proto)(gp_in->proto,
					  &(gp_out->proto));
    }
    if (gp_in->all_protos)
    { 
	if (gp_in->all_protos == my_mi_protos)
	{
	    gp_out->all_protos = my_mi_protos;
	}
	else
	{
	    _nx_copy_mi_proto( gp_in->all_protos,
			       &(gp_out->all_protos) );
	}
    } else {
	gp_out->all_protos = gp_in->all_protos; /* NULL */
    }
} /* nexus_copy_global_pointer() */

/* 
 * _nx_copy_mi_proto()
 *
 * Copies the nexus_mi_proto_t structure and its associated strings
 */
static void _nx_copy_mi_proto( nexus_mi_proto_t *mi_proto_in,
			       nexus_mi_proto_t **mi_proto_out)
{
    int size,j;
    nexus_mi_proto_t *mi_proto;
    nexus_mi_proto_t *new_mi_proto;
    nexus_mi_proto_t **cur_mi_proto;

    cur_mi_proto = mi_proto_out;
    *mi_proto_out = NULL;

    for (mi_proto = mi_proto_in; mi_proto; mi_proto = mi_proto->next)
    {
	/* allocate room for the nexus_mi_proto_t */
	size = (sizeof(nexus_mi_proto_t)		/* base */
		+sizeof(int)*(mi_proto->n_ints)	/* for ints */
		+sizeof(int)*(mi_proto->n_strings) /* for string_lengths*/
		+sizeof(char *)*(mi_proto->n_strings));/*for string_ptrs*/
	NexusMalloc(_nx_copy_mi_proto(),
		    new_mi_proto,
		    nexus_mi_proto_t *,
		    size);
	new_mi_proto->proto_type = mi_proto->proto_type;
	new_mi_proto->n_ints = mi_proto->n_ints;
	new_mi_proto->n_strings = mi_proto->n_strings;
	
	/* copy the ints */
	if (mi_proto->n_ints > 0)
	{
	    new_mi_proto->ints = (int *) (((char *) new_mi_proto)
					     + sizeof(nexus_mi_proto_t));
	    
	    for (j = 0; j < mi_proto->n_ints; j++)
	    {
		new_mi_proto->ints[j] = mi_proto->ints[j];
	    }
	}
	
	/* copy the string_lengths */
	if (mi_proto->n_strings > 0)
	{
	    new_mi_proto->string_lengths
		= (int *) (((char *) new_mi_proto)
			   + sizeof(nexus_mi_proto_t)
			   + sizeof(int) * new_mi_proto->n_ints);
	    new_mi_proto->strings
		= (char **) (((char *) new_mi_proto)
			     + sizeof(nexus_mi_proto_t)
			     + sizeof(int) * new_mi_proto->n_ints
			     + sizeof(int) * new_mi_proto->n_strings);
	    for (j = 0; j < mi_proto->n_strings; j++)
	    {
		new_mi_proto->string_lengths[j] = 
		    mi_proto->string_lengths[j];
	    }
    
	    /* copy the strings */
	    for (j = 0; j < mi_proto->n_strings; j++)
	    {
		new_mi_proto->strings[j] = 
		    _nx_copy_string(mi_proto->strings[j]);
	    }
	}
	/*
	 * tack new_mi_proto onto the mi_proto_out list
	 */
	*cur_mi_proto = mi_proto;
	cur_mi_proto = &((*cur_mi_proto)->next);
	*cur_mi_proto = NULL;
    }
} /* _nx_copy_mi_proto() */

/*
 * nexus_sizeof_global_pointer()
 *
 * Return the size in bytes needed to encode the passed
 * global pointer, 'gp', into the 'buffer'.
 * Also, set 'n_elements' to the number of puts that will be
 * done when nexus_put_global_pointer() is called for this gp.
 */
int nexus_sizeof_global_pointer(nexus_buffer_t *buffer,
				nexus_global_pointer_t *gp,
				int count,
				int *n_elements)
{
    nexus_global_pointer_t *g;
    nexus_mi_proto_t *mi_proto;
    int j;
    int c;
    int size = 0;
    int n = 0;

    NexusAssert2((gp != (nexus_global_pointer_t *) NULL),
		 ("nexus_sizeof_global_pointer(): Passed a NULL pointer to a gp.\n"));
		
    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	/* sizeof the context and address */
	size += nexus_sizeof_u_long(buffer, 2);
	size += nexus_sizeof_u_long(buffer, 1);
	n++;
	
#ifdef NEXUS_PROFILE
	/* sizeof the node_id, context_id */
	size += nexus_sizeof_int(buffer, 2);
	n++;
#endif /* NEXUS_PROFILE */    
	
	/* sizeof the n_protos */
	size += nexus_sizeof_int(buffer, 1);
	n++;
	
	for (mi_proto = g->all_protos; mi_proto; mi_proto = mi_proto->next)
	{
	    /* sizeof the proto_type, n_ints, and n_strings */
	    size += nexus_sizeof_int(buffer, 3);
	    n++;

	    if (mi_proto->n_ints > 0)
	    {
		/* sizeof the ints */
		size += nexus_sizeof_int(buffer, mi_proto->n_ints);
		n++;
	    }

	    if (mi_proto->n_strings > 0)
	    {
		/* sizeof the string_lengths */
		size += nexus_sizeof_int(buffer, mi_proto->n_strings);
		n++;
		
		/* sizeof the strings */
		for (j = 0; j < mi_proto->n_strings; j++)
		{
		    size += nexus_sizeof_char(buffer,
					      mi_proto->string_lengths[j]);
		    n++;
		}
	    }
	}
    }
    
    if (n_elements != (int *) NULL)
	*n_elements = n;
    return (size);
} /* nexus_sizeof_global_pointer() */


/*
 * nexus_put_global_pointer()
 *
 * Signal a NULL global pointer by making n_protos = -1
 */
void nexus_put_global_pointer(nexus_buffer_t *buffer,
			      nexus_global_pointer_t *gp,
			      int count)
{
    nexus_global_pointer_t *g;
    nexus_mi_proto_t *mi_proto;
    int j;
    int c;
    int n_protos;

    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	/* check that everything is ok first */
	NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "nexus_put_global_pointer");
	/* put the context, address */
	nexus_put_u_long(buffer, &(g->context), 2);
	nexus_put_u_long(buffer, (unsigned long *)(&(g->class_type)), 1);
	
#ifdef NEXUS_PROFILE
	/* put the node_id, context_id */
	nexus_put_int(buffer, &(g->node_id), 2);
#endif /* NEXUS_PROFILE */    

	/* put n_protos */
	if (g->proto == (nexus_proto_t *) NULL)
	{
	    n_protos = -1;
	}
	else
	{
	    for (mi_proto = g->all_protos, n_protos = 0;
		 mi_proto;
		 mi_proto = mi_proto->next, n_protos++)
		;
	}
	nexus_put_int(buffer, &n_protos, 1);
	
	for (mi_proto = g->all_protos; mi_proto; mi_proto = mi_proto->next)
	{
	    /* put the proto_type, n_ints, and n_strings */
	    nexus_put_int(buffer, &(mi_proto->proto_type), 3);

	    if (mi_proto->n_ints > 0)
	    {
		/* put the ints */
		nexus_put_int(buffer, mi_proto->ints, mi_proto->n_ints);
	    }

	    if (mi_proto->n_strings > 0)
	    {
		/* put the string_lengths */
		nexus_put_int(buffer, mi_proto->string_lengths,
			      mi_proto->n_strings);
		
		/* put the strings */
		for (j = 0; j < mi_proto->n_strings; j++)
		{
		    nexus_put_char(buffer, mi_proto->strings[j],
				   mi_proto->string_lengths[j]);
		}
	    }
	}
    }
} /* nexus_put_global_pointer() */


/*
 * nexus_get_global_pointer()
 */
void nexus_get_global_pointer(nexus_buffer_t *buffer,
			      nexus_global_pointer_t *gp,
			      int count)
{
    nexus_global_pointer_t *g;
    nexus_mi_proto_t tmp_mi_proto;
    nexus_mi_proto_t *mi_proto, *last_mi_proto;
    int i, j, l;
    int c, size;
    int n_protos;
    char *s;
    proto_module_list_t *proto_module;

    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	g->proto = (nexus_proto_t *) NULL;
	g->all_protos = (nexus_mi_proto_t *) NULL;

	/* get the context, address */
	nexus_get_u_long(buffer, &(g->context), 2);
	nexus_get_u_long(buffer, (unsigned long *)(&(g->class_type)), 1);
	
#ifdef NEXUS_PROFILE
	/* get the node_id, context_id */
	nexus_get_int(buffer, &(g->node_id), 2);
#endif /* NEXUS_PROFILE */    

	/* get n_protos */
	nexus_get_int(buffer, &n_protos, 1);

	if (n_protos == 0)
	{
	    g->proto = _nx_local_proto;
	}
	/* n_protos<0 means its a NULL global pointer */
	  
	last_mi_proto = (nexus_mi_proto_t *) NULL;
	for (i = 0; i < n_protos; i++)
	{
	    /* get the proto_type, n_ints, and n_strings */
	    nexus_get_int(buffer, &(tmp_mi_proto.proto_type), 3);
	    
	    /* allocate room for the nexus_mi_proto_t */
	    size = (sizeof(nexus_mi_proto_t)		/* base */
		    +sizeof(int)*tmp_mi_proto.n_ints	/* for ints */
		    +sizeof(int)*tmp_mi_proto.n_strings /* for string_lengths*/
		    +sizeof(char *)*tmp_mi_proto.n_strings);/*for string_ptrs*/
	    NexusMalloc(nexus_get_global_pointer(),
			mi_proto,
			nexus_mi_proto_t *,
			size);
	    mi_proto->proto_type = tmp_mi_proto.proto_type;
	    mi_proto->n_ints = tmp_mi_proto.n_ints;
	    mi_proto->n_strings = tmp_mi_proto.n_strings;
	    
	    /* get the ints */
	    if (mi_proto->n_ints > 0)
	    {
		mi_proto->ints = (int *) (((char *) mi_proto)
					  + sizeof(nexus_mi_proto_t));
		nexus_get_int(buffer, mi_proto->ints,
			      mi_proto->n_ints);
	    }
	    
	    /* get the string_lengths */
	    if (mi_proto->n_strings > 0)
	    {
		mi_proto->string_lengths
		    = (int *) (((char *) mi_proto)
			       + sizeof(nexus_mi_proto_t)
			       + sizeof(int) * mi_proto->n_ints);
		mi_proto->strings
		    = (char **) (((char *) mi_proto)
				 + sizeof(nexus_mi_proto_t)
				 + sizeof(int) * mi_proto->n_ints
				 + sizeof(int) * mi_proto->n_strings);
		nexus_get_int(buffer, mi_proto->string_lengths,
			      mi_proto->n_strings);
		for (j = 0; j < mi_proto->n_strings; j++)
		{
		    l = mi_proto->string_lengths[j];
		    NexusMalloc(nexus_get_global_pointer(),
				s, char *, (l + 1));
		    mi_proto->strings[j] = s;
		    nexus_get_char(buffer, s, l);
		    s[l] = '\0';
		}
	    }
	    
	    /*
	     * Add this mi_proto to the all_protos list
	     */
	    if (last_mi_proto == (nexus_mi_proto_t *) NULL)
	    {
		g->all_protos = last_mi_proto = mi_proto;
	    }
	    else
	    {
		last_mi_proto->next = mi_proto;
		last_mi_proto = mi_proto;
	    }
	    mi_proto->next = (nexus_mi_proto_t *) NULL;
	    
	    /*
	     * I now have a complete mi_proto.
	     * If I haven't found a proto that I can use yet, then
	     * search my protocol list to see if I can support this
	     * protocol.  If so, then call down into the appropriate
	     * protocol module to unpack it for use.
	     *
	     * If construct_from_mi_proto() recognizes that this is
	     * a local proto, then it can fill in g->proto
	     * with _nx_local_proto.
	     *
	     * Note: We don't have to check against the local
	     * proto (i.e., we start at k=1) since the
	     * local proto is not passed around in the mi_proto list.
	     */
	    if (g->proto == (nexus_proto_t *) NULL)
	    {
		for (proto_module = proto_module_list_head;
		     proto_module;
		     proto_module = proto_module->next)
		{
		    if (proto_module->type == mi_proto->proto_type)
		    {
			(proto_module->funcs->construct_from_mi_proto)(
								&(g->proto),
								mi_proto);
		    }
		}
	    }
	}
	NexusAssert2((   (g->proto != (nexus_proto_t *) NULL)
		      || (g->all_protos == (nexus_mi_proto_t *) NULL) ),
		     ("nexus_get_global_pointer(): Could not find a usable protocol.\n"));
	NEXUS_INTERROGATE(g, _NX_GLOBAL_POINTER_T, "nexus_get_global_pointer");	
    }
	
} /* nexus_get_global_pointer() */


/*
 * nexus_get_stashed_global_pointer()
 *
 * This is identical to nexus_get_global_pointer(), except that
 * nexus_get_stashed_TYPE() is used instead of nexus_get_TYPE().
 */
void nexus_get_stashed_global_pointer(nexus_stashed_buffer_t *buffer,
				      nexus_global_pointer_t *gp,
				      int count)
{
    nexus_global_pointer_t *g;
    nexus_mi_proto_t tmp_mi_proto;
    nexus_mi_proto_t *mi_proto, *last_mi_proto;
    int i, j, l;
    int c, size;
    int n_protos;
    char *s;
    proto_module_list_t *proto_module;

    for (c = 0, g = gp;
	 c < count;
	 c++, g++)
    {
	g->proto = (nexus_proto_t *) NULL;
	g->all_protos = (nexus_mi_proto_t *) NULL;

	/* get the context, address */
	nexus_get_stashed_u_long(buffer, &(g->context), 2);
	nexus_get_stashed_u_long(buffer, (unsigned long *)(&(g->class_type)), 1);
	
#ifdef NEXUS_PROFILE
	/* get the node_id, context_id */
	nexus_get_stashed_int(buffer, &(g->node_id), 2);
#endif /* NEXUS_PROFILE */    

	/* get n_protos */
	nexus_get_stashed_int(buffer, &n_protos, 1);

	if (n_protos == 0)
	{
	    g->proto = _nx_local_proto;
	}
	/* n_protos<0 means its a NULL global pointer */

	last_mi_proto = (nexus_mi_proto_t *) NULL;
	for (i = 0; i < n_protos; i++)
	{
	    /* get the proto_type, n_ints, and n_strings */
	    nexus_get_stashed_int(buffer, &(tmp_mi_proto.proto_type), 3);
	    
	    /* allocate room for the nexus_mi_proto_t */
	    size = (sizeof(nexus_mi_proto_t)		/* base */
		    +sizeof(int)*tmp_mi_proto.n_ints	/* for ints */
		    +sizeof(int)*tmp_mi_proto.n_strings /* for string_lengths*/
		    +sizeof(char *)*tmp_mi_proto.n_strings);/*for string_ptrs*/
	    NexusMalloc(nexus_get_stashed_global_pointer(),
			mi_proto,
			nexus_mi_proto_t *,
			size);
	    mi_proto->proto_type = tmp_mi_proto.proto_type;
	    mi_proto->n_ints = tmp_mi_proto.n_ints;
	    mi_proto->n_strings = tmp_mi_proto.n_strings;
	    
	    /* get the ints */
	    if (mi_proto->n_ints > 0)
	    {
		mi_proto->ints = (int *) (((char *) mi_proto)
					  + sizeof(nexus_mi_proto_t));
		nexus_get_stashed_int(buffer, mi_proto->ints,
			      mi_proto->n_ints);
	    }
	    
	    /* get the string_lengths */
	    if (mi_proto->n_strings > 0)
	    {
		mi_proto->string_lengths
		    = (int *) (((char *) mi_proto)
			       + sizeof(nexus_mi_proto_t)
			       + sizeof(int) * mi_proto->n_ints);
		mi_proto->strings
		    = (char **) (((char *) mi_proto)
				 + sizeof(nexus_mi_proto_t)
				 + sizeof(int) * mi_proto->n_ints
				 + sizeof(int) * mi_proto->n_strings);
		nexus_get_stashed_int(buffer, mi_proto->string_lengths,
			      mi_proto->n_strings);
		for (j = 0; j < mi_proto->n_strings; j++)
		{
		    l = mi_proto->string_lengths[j];
		    NexusMalloc(nexus_get_stashed_global_pointer(),
				s, char *, (l + 1));
		    mi_proto->strings[j] = s;
		    nexus_get_stashed_char(buffer, s, l);
		    s[l] = '\0';
		}
	    }
	    
	    /*
	     * Add this mi_proto to the all_protos list
	     */
	    if (last_mi_proto == (nexus_mi_proto_t *) NULL)
	    {
		g->all_protos = last_mi_proto = mi_proto;
	    }
	    else
	    {
		last_mi_proto->next = mi_proto;
		last_mi_proto = mi_proto;
	    }
	    mi_proto->next = (nexus_mi_proto_t *) NULL;
	    
	    /*
	     * I now have a complete mi_proto.
	     * If I haven't found a proto that I can use yet, then
	     * search my protocol list to see if I can support this
	     * protocol.  If so, then call down into the appropriate
	     * protocol module to unpack it for use.
	     *
	     * If construct_from_mi_proto() recognizes that this is
	     * a local proto, then it can fill in g->proto
	     * with _nx_local_proto.
	     *
	     * Note: We don't have to check against the local
	     * proto (i.e., we start at k=1) since the
	     * local proto is not passed around in the mi_proto list.
	     */
	    if (g->proto == (nexus_proto_t *) NULL)
	    {
		for (proto_module = proto_module_list_head;
		     proto_module;
		     proto_module = proto_module->next)
		{
		    if (proto_module->type == mi_proto->proto_type)
		    {
			(proto_module->funcs->construct_from_mi_proto)(
								&(g->proto),
								mi_proto);
		    }
		}
	    }
	}
	NexusAssert2((   (g->proto != (nexus_proto_t *) NULL)
		      || (g->all_protos == (nexus_mi_proto_t *) NULL) ),
		     ("nexus_get_stashed_global_pointer(): Could not find a usable protocol.\n"));
	NEXUS_INTERROGATE(g, _NX_GLOBAL_POINTER_T, "nexus_get_stashed_global_pointer");

    }
	
} /* nexus_get_stashed_global_pointer() */


/*
 * _nx_write_global_pointer()
 *
 * Write the global pointer to the passed file descriptor.
 */
void _nx_write_global_pointer(int filedesc, nexus_global_pointer_t *gp)
{
    int n_protos;
    nexus_mi_proto_t *mi_proto;
    int j;

    NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "_nx_write_global_pointer");


    /* put the context, address */
    _nx_write_u_long(filedesc, (unsigned long) gp->context);
    _nx_write_u_long(filedesc, (unsigned long) gp->address);
    _nx_write_u_long(filedesc, (unsigned long) gp->class_type);

#ifdef NEXUS_PROFILE
    /* put the node_id, context_id */
    _nx_write_u_long(filedesc, (unsigned long) gp->node_id);
    _nx_write_u_long(filedesc, (unsigned long) gp->context_id);
#endif /* NEXUS_PROFILE */    

    /* put n_protos */
    if (gp->proto == (nexus_proto_t *) NULL)
    {
	n_protos = -1;
    }
    else
    {
	for( mi_proto = gp->all_protos, n_protos = 0;
	    mi_proto;
	    mi_proto = mi_proto->next, n_protos++ )
	    ;
    }
    _nx_write_u_long(filedesc, (unsigned long) n_protos);
    
    for (mi_proto = gp->all_protos; mi_proto; mi_proto = mi_proto->next )
    {
	/* put the proto_type, n_ints, and n_strings */
	_nx_write_u_long(filedesc, (unsigned long) mi_proto->proto_type);
	_nx_write_u_long(filedesc, (unsigned long) mi_proto->n_ints);
	_nx_write_u_long(filedesc, (unsigned long) mi_proto->n_strings);

	if( mi_proto->n_ints > 0 )
	{
	    /* put the ints */
	    for (j = 0; j < mi_proto->n_ints; j++)
	    {
		_nx_write_u_long(filedesc,
				 (unsigned long) mi_proto->ints[j]);
	    }
	}
	if( mi_proto->n_strings > 0 )
	{
	    /* put the string_lengths */
	    for (j = 0; j < mi_proto->n_strings; j++)
	    {
		_nx_write_u_long(filedesc,
				 (unsigned long) mi_proto->string_lengths[j]);
	    }
	    
	    /* put the strings */
	    for (j = 0; j < mi_proto->n_strings; j++)
	    {
		_nx_write(filedesc,
			  (char *) mi_proto->strings[j],
			  mi_proto->string_lengths[j] );
	    }
	}
    }
} /* _nx_write_global_pointer() */


/*
 * _nx_read_global_pointer()
 *
 * Read a global pointer from the passed file descriptor and fill
 * it into 'gp'.
 */
void _nx_read_global_pointer(int filedesc, nexus_global_pointer_t *gp)
{
    int i,j,l;
    int size;
    nexus_mi_proto_t tmp_mi_proto;
    nexus_mi_proto_t *mi_proto, *last_mi_proto;
    int n_protos;
    char *s;
    proto_module_list_t *proto_module;

    gp->proto = (nexus_proto_t *) NULL;
    gp->all_protos = (nexus_mi_proto_t *) NULL;

    /* get the context, address */
    gp->context = _nx_read_u_long(filedesc);
    gp->address = _nx_read_u_long(filedesc);
    gp->class_type = _nx_read_u_long(filedesc);

#ifdef NEXUS_PROFILE
    /* get the node_id, context_id */
    gp->node_id = _nx_read_u_long(filedesc);
    gp->context_id = _nx_read_u_long(filedesc);
#endif /* NEXUS_PROFILE */    

    /* get n_protos */
    n_protos = _nx_read_u_long(filedesc);

    if (n_protos == 0)
    {
	gp->proto = _nx_local_proto;
    }
    /* n_protos<0 means its a NULL global pointer */

    last_mi_proto = (nexus_mi_proto_t *) NULL;
    for (i = 0; i < n_protos; i++)
    {
	/* get the proto_type, n_ints, and n_strings */
	tmp_mi_proto.proto_type = _nx_read_u_long(filedesc);
	tmp_mi_proto.n_ints = _nx_read_u_long(filedesc);
	tmp_mi_proto.n_strings = _nx_read_u_long(filedesc);

	/* allocate room for the nexus_mi_proto_t */
	size = (sizeof(nexus_mi_proto_t)		/* base */
		+sizeof(int)*tmp_mi_proto.n_ints	/* for ints */
		+sizeof(int)*tmp_mi_proto.n_strings	/* for string_lengths*/
		+sizeof(char *)*tmp_mi_proto.n_strings);/* for string_ptrs*/
	NexusMalloc(_nx_read_global_pointer(),
		    mi_proto,
		    nexus_mi_proto_t *,
		    size);
	mi_proto->proto_type = tmp_mi_proto.proto_type;
	mi_proto->n_ints = tmp_mi_proto.n_ints;
	mi_proto->n_strings = tmp_mi_proto.n_strings;

	/* get the ints */
	if (mi_proto->n_ints > 0)
	{
	    mi_proto->ints = (int *) (((char *) mi_proto)
				      + sizeof(nexus_mi_proto_t));
	    for (j = 0; j < mi_proto->n_ints; j++)
	    {
		mi_proto->ints[j] = _nx_read_u_long(filedesc);
	    }
	}
	
	/* get the string_lengths */
	if (mi_proto->n_strings > 0)
	{
	    mi_proto->string_lengths
		= (int *) (((char *) mi_proto)
			   + sizeof(nexus_mi_proto_t)
			   + sizeof(int) * mi_proto->n_ints);
	    mi_proto->strings
		= (char **) (((char *) mi_proto)
			     + sizeof(nexus_mi_proto_t)
			     + sizeof(int) * mi_proto->n_ints
			     + sizeof(int) * mi_proto->n_strings);
	    for (j = 0; j < mi_proto->n_strings; j++)
	    {
		mi_proto->string_lengths[j] = _nx_read_u_long(filedesc);
	    }
	    for (j = 0; j < mi_proto->n_strings; j++)
	    {
		l = mi_proto->string_lengths[j];
		NexusMalloc(_nx_read_global_pointer(),
			    s, char *, (l + 1));
		mi_proto->strings[j] = s;
		_nx_read(filedesc, s, l);
		s[l] = '\0';
	    }
	}
	
	/*
	 * Add this mi_proto to the all_protos list
	 */
	if(last_mi_proto == (nexus_mi_proto_t *) NULL)
	{
	    gp->all_protos = last_mi_proto = mi_proto;
	}
	else
	{
	    last_mi_proto->next = mi_proto;
	    last_mi_proto = mi_proto;
	}
	mi_proto->next = (nexus_mi_proto_t *) NULL;
	
	/*
	 * I now have a complete mi_proto.
	 * If I haven't found a proto that I can use yet, then
	 * search my protocol list to see if I can support this
	 * protocol.  If so, then call down into the appropriate
	 * protocol module to unpack it for use.
	 *
	 * If construct_from_mi_proto() recognizes that this is
	 * a local proto, then it can fill in g->proto
	 * with _nx_local_proto.
	 *
	 * Note: We don't have to check against the local
	 * proto (i.e., we start at k=1) since the
	 * local proto is not passed around in the mi_proto list.
	 */
	if (gp->proto == (nexus_proto_t *) NULL)
	{
	    for (proto_module = proto_module_list_head;
		 proto_module;
		 proto_module = proto_module->next)
	    {
		if (proto_module->type == mi_proto->proto_type)
		{
		    (proto_module->funcs->construct_from_mi_proto)(
								&(gp->proto),
								mi_proto);
		}
	    }
	}
    }
    NexusAssert2((   (gp->proto != (nexus_proto_t *) NULL)
		  || (gp->all_protos == (nexus_mi_proto_t *) NULL) ),
		 ("_nx_read_global_pointer(): Could not find a usable protocol.\n"));
    NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "_nx_read_global_pointer");

} /* _nx_read_global_pointer() */


#ifdef NEXUS_DEBUG
/*
 * nexus_global_pointer_string()
 *
 * Return a string representation of the global pointer.
 * The returned string does not need to be freed, coming as it does
 * from a pool of static buffers. Note we lock in here.
 */
char *nexus_global_pointer_string(nexus_global_pointer_t *g)
{
    static char strings[GP_N_STRINGS][1024];
    static int current_string = 0;
    nexus_mi_proto_t *mi_proto;
    int j;
    int n_protos;
    char *buf;
    char temp1[50];
    int i;

    nexus_mutex_lock(&gp_string_mutex);
    buf = strings[current_string];
    current_string = ((current_string + 1) % GP_N_STRINGS);
    nexus_mutex_unlock(&gp_string_mutex);
    
    if (g->proto == (nexus_proto_t *) NULL)
    {
        n_protos = -1;
    }
    else
    {
        for (mi_proto = g->all_protos, n_protos = 0;
             mi_proto;
             mi_proto = mi_proto->next, n_protos++)
            ;
    }

    buf[0] = '\0';

#ifdef NEXUS_PROFILE
    sprintf(buf, "[GP:%lx:%lx:%lx:%d:n%d:c%d",
	    (unsigned long) g,
	    (unsigned long) g->context,
	    (unsigned long) g->address,
	    n_protos,
	    g->node_id,
	    g->context_id);
#else  /* NEXUS_PROFILE */    
    sprintf(buf, "[GP:%lx:%lx:%lx:%d:",
	    (unsigned long) g,
	    (unsigned long) g->context,
	    (unsigned long) g->address,
	    n_protos);
#endif /* NEXUS_PROFILE */    
              
    for (mi_proto = g->all_protos; mi_proto; mi_proto = mi_proto->next)
    {
        strcat(buf, "<{");

        for (i = 0; i < mi_proto->n_ints; i++)
        {
            sprintf(temp1, "%d", mi_proto->ints[i]);
            strcat(buf, temp1);
            if (i < mi_proto->n_ints - 1)
                strcat(buf, "," );
        }
        strcat(buf, "}{");

        /* put the strings */
        for (j = 0; j < mi_proto->n_strings; j++)
        {
            strcat(buf, "\"");
            strncat(buf, mi_proto->strings[j], mi_proto->string_lengths[j]);
            strcat(buf, "\"");
            if (j < mi_proto->n_strings - 1)
                strcat(buf, ",");
        }
        strcat(buf, "}>");
    }
    strcat(buf, "]");
    return buf;
} /* nexus_global_pointer_string() */
#endif /* NEXUS_DEBUG */


/*
 * free_mi_proto()
 */
static void free_mi_proto(nexus_mi_proto_t *mi_proto)
{
    int i;
    for (i = 0; i < mi_proto->n_strings; i++)
    {
	NexusFree(mi_proto->strings[i]);
    }
    NexusFree(mi_proto);
} /* free_mi_proto() */


/*
 * nexus_same_context()
 */
int nexus_same_context(nexus_global_pointer_t *gp1,
		       nexus_global_pointer_t *gp2)
{
    NEXUS_INTERROGATE(gp1, _NX_GLOBAL_POINTER_T, "nexus_same_context");
    NEXUS_INTERROGATE(gp2, _NX_GLOBAL_POINTER_T, "nexus_same_context");

    if (   (gp1->proto == (nexus_proto_t *) NULL)
	&& (gp2->proto == (nexus_proto_t *) NULL) )
    {
	/* Both are NULL global pointers */
	return (NEXUS_TRUE);
    }
    if (   (gp1->proto == (nexus_proto_t *) NULL)
	|| (gp2->proto == (nexus_proto_t *) NULL) )
    {
	/* One is a NULL global pointer, the other isn't */
	return (NEXUS_FALSE);
    }
    if (gp1->context != gp2->context)
    {
	return (NEXUS_FALSE);
    }
    if (gp1->proto->type != gp2->proto->type)
    {
	return (NEXUS_FALSE);
    }
    /*
     * Both gp's have the save type proto,
     * so call down into proto to compare them.
     */
    return ((gp1->proto->funcs->compare_protos)(gp1->proto, gp2->proto));
} /* nexus_same_context() */


/*
 * nexus_same_global_pointer()
 */
int nexus_same_global_pointer(nexus_global_pointer_t *gp1,
			      nexus_global_pointer_t *gp2)
{
    return (   (nexus_same_context(gp1, gp2))
	    && (gp1->address == gp2->address) );
} /* nexus_same_global_pointer() */


/*
 * nexus_global_pointer_to_current_context()
 */
int nexus_global_pointer_to_current_context(nexus_global_pointer_t *gp)
{
    nexus_context_t *context;

    NEXUS_INTERROGATE(gp, _NX_GLOBAL_POINTER_T, "nexus_global_pointer_to_current_context");

    if (   (gp->proto != (nexus_proto_t *) NULL)
	&& (gp->proto->type == NEXUS_PROTO_TYPE_LOCAL) )
    {
	_nx_context(&context);
	if (gp->context == (unsigned long) context)
	    return (NEXUS_TRUE);
	else
	    return (NEXUS_FALSE);
    }
    else
    {
	return (NEXUS_FALSE);
    }
} /* nexus_global_pointer_to_current_context() */


/*
 * _nx_proto_get_creator_gp_params()
 *
 * Fill in 'buf', up to 'buf_size' bytes, with the parameters
 * necessary for a new process to connect to this one.
 *
 * If 'proto_module_string' is not NULL, use that protocol module
 * for the initial connection.  Otherwise, use the 'node_name' and
 * 'node_number' to lookup candidate protocol modules for the new
 * node and find a protocol module to use by comparing that
 * candidate list against the protocols that this process knows.
 *
 * The 'context' and 'address' are for the global pointer that
 * the new process should use to checkin.
 */
void _nx_proto_get_creator_gp_params(char *buf,
				     int buf_size,
				     char *node_name,
				     int node_number,
				     char *proto_module_string,
				     unsigned long context,
				     unsigned long address,
				     unsigned long class_type)
{
    nexus_bool_t done;
    char *value, *next;
    proto_module_list_t *proto_module;
    int len;
#ifdef NEXUS_PROFILE
    char tmp_buf[64];
    int my_node_id, my_context_id;
#endif /* NEXUS_PROFILE */

    /*
     * Figure out which protocol to use for the node.
     * Get the list or protocols that the node supports,
     * find the first one in that list that this node also
     * supports, and load that protocol module if necessary.
     *
     * TODO: This is hacked to just use "tcp" protocol module
     *       if a 'proto_module_string' is not supplied.		
     *       Fix it to use _nx_database_lookup().
     */
    if (proto_module_string == (char *) NULL)
    {
	proto_module_string = nexus_database_lookup(node_name,
						    node_number,
						    "protocols");
	if (!proto_module_string) 
	{
	    /* Default value */
	    proto_module_string = _nx_copy_string("tcp");
	}
    }
    done = NEXUS_FALSE;
    _nx_get_next_value(proto_module_string, ',', &next, &value);
    while (!done && value)
    {
        for (proto_module = proto_module_list_head;
	     proto_module;
	     proto_module = proto_module->next)
        {
	    if (strcmp(proto_module->name, value) == 0)
	    {
	        done = NEXUS_TRUE;
	        break;
	    }
        }
        if (done)
        {
	    break;
        }
	if (next)
	{
            _nx_get_next_value(next, ',', &next, &value);
	}
    }
    if (!done)
    {
	nexus_fatal("_nx_proto_get_creator_gp_params(): Cannot find %s protocol module\n", proto_module_string);
    }

    /*
     * Put parameters into 'buf' to specify:
     *   -nf_nx_prmod <module> : which protocol module to use
     *   -nf_nx_gp <context> <address> : global pointer info
     *   -nf_nx_gpid <node_id> <context_id> : Add this if profiling 
     */
    nexus_stdio_lock();
    sprintf(buf, "-nf_nx_prmod %s -nf_nx_gp \"%lu %lu %lu\" ",
	    proto_module->name,
	    context,
	    address,
	    class_type);
#ifdef NEXUS_PROFILE
    _nx_node_id(&my_node_id);
    _nx_context_id(&my_context_id);
    sprintf(tmp_buf, "-nf_nx_gpid \"%d %d\" ", my_node_id, my_context_id);
    if (strlen(buf) + strlen(tmp_buf) > (size_t)buf_size)
    {
	nexus_fatal("_nx_proto_get_creator_gp_params(): Internal error: Buffer not big enough for arguments\n");
    }
    strcat(buf, tmp_buf);
#endif /* NEXUS_PROFILE */
    nexus_stdio_unlock();

    /*
     * Now add the parameters for the protocol
     */
    len = strlen(buf);
    (*proto_module->funcs->get_creator_proto_params)(buf + len,
						     buf_size - len,
						     node_name,
						     node_number);
    
} /* _nx_proto_get_creator_gp_params */


/*
 * _nx_proto_construct_creator_gp()
 */
void _nx_proto_construct_creator_gp(nexus_global_pointer_t *creator_gp)
{
    char *proto_name;
    proto_module_list_t *proto_module;
    char *gp_string;
    unsigned long context;
    unsigned long address;
    unsigned long class_type;
    int rc;
    int arg_num;
    int *argc;
    char ***argv;
#ifdef NEXUS_PROFILE
    char *gpid_string;
    int node_id;
    int context_id;
#endif /* NEXUS_PROFILE */				    

    nexus_get_retained_arguments(&argc, &argv);
    
    /*
     * Figure out which proto to use
     *
     * TODO: To make this more dynamic, the protocol module may need
     * to be dynamically loaded here.
     */
    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_prmod", 2)) >= 0)
    {
	proto_name = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Expected a -nf_nx_prmod argument.\n");
    }
    for (proto_module = proto_module_list_head;
	 proto_module;
	 proto_module = proto_module->next)
    {
	if (strcmp(proto_module->name, proto_name) == 0)
	{
	    break;
	}
    }
    if (!proto_module)
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): I do not have the \"%s\" protocol module, so cannot checkin.\n", proto_name);
    }
    
    /*
     * Get the global pointer context and address
     */
    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_gp", 2)) >= 0)
    {
	gp_string = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Expected a -nf_nx_gp argument.\n");
    }
    nexus_stdio_lock();
    rc = sscanf(gp_string, "%lu %lu %lu", &context, &address, &class_type);
    nexus_stdio_unlock();
    if (rc != 3)
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Invalid -Rnx_gp argument.\n");
    }

#ifdef NEXUS_PROFILE    
    /*
     * Get the global pointer context and address
     */
    if ((arg_num = nexus_find_argument(argc, argv, "nf_nx_gpid", 2)) >= 0)
    {
	gpid_string = (*argv)[arg_num + 1];
	nexus_remove_arguments(argc, argv, arg_num, 2);
    }
    else
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Expected a -nf_nx_gpid argument.\n");
    }
    nexus_stdio_lock();
    rc = sscanf(gpid_string, "%d %d", &node_id, &context_id);
    nexus_stdio_unlock();
    if (rc != 2)
    {
	nexus_fatal("_nx_proto_construct_creator_gp(): Internal error: Invalid -Rnx_gpid argument.\n");
    }
#endif /* NEXUS_PROFILE */
    
    (*proto_module->funcs->construct_creator_proto)(&(creator_gp->proto));
    creator_gp->all_protos = (nexus_mi_proto_t *) NULL;
    creator_gp->context = context;
    creator_gp->address = address;
    creator_gp->class_type = class_type;
#ifdef NEXUS_PROFILE
    creator_gp->node_id = node_id;
    creator_gp->context_id = context_id;
#endif /* NEXUS_PROFILE */    

} /* _nx_proto_construct_creator_gp() */


/*
 * _nx_write()
 *
 * This is a wrapper around write() which will keep trying until
 * 'len' characters have been written.  This will work on either
 * blocking or non-blocking file descriptors, and will properly
 * retry on EINTR.
 */
int _nx_write(int fd, char *buf, int len)
{
    char *b = buf;
    int l = len;
    int rc;
    while (l > 0)
    {
	rc = write(fd, b, l);
	if (rc >= 0)
	{
	    b += rc;
	    l -= rc;
	}
	else
	{
	    if (errno == EINTR || errno == EWOULDBLOCK)
	    {
		/* Do nothing.  Just try again. */
	    }
	    else
	    {
		nexus_fatal("_nx_write(): Write failed: %s\n",
			    _nx_md_system_error_string());
	    }
	}
    }
    return(len);
} /* _nx_write() */


/*
 * _nx_read()
 *
 * This is a wrapper around read() which will keep trying until
 * 'len' characters have been read.  This will work on either
 * blocking or non-blocking file descriptors, and will properly
 * retry on EINTR.
 */
int _nx_read(int fd, char *buf, int len)
{
    char *b = buf;
    int l = len;
    int rc;
    while (l > 0)
    {
	rc = read(fd, b, l);
	if (rc >= 0)
	{
	    b += rc;
	    l -= rc;
	}
	else
	{
	    if (errno == EINTR || errno == EWOULDBLOCK)
	    {
		/* Do nothing.  Just try again. */
	    }
	    else
	    {
		nexus_fatal("_nx_read(): Read failed: %s\n",
			    _nx_md_system_error_string());
	    }
	}
    }
    return(len);
} /* _nx_read() */


/*
 * _nx_write_u_long()
 */
void _nx_write_u_long(int fd, unsigned long l)
{
    unsigned long tmp;
    tmp = NX_HTONL(l);
    _nx_write(fd, (char *) &tmp, sizeof(unsigned long));
} /* _nx_write_u_long() */


/*
 * _nx_read_u_long()
 */
unsigned long _nx_read_u_long(int fd)
{
    unsigned long tmp;
    _nx_read(fd, (char *) &tmp, sizeof(unsigned long));
    tmp = NX_NTOHL(tmp);
    return(tmp);
} /* _nx_read_u_long() */
